/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 *  
 *  2/27/00:3am, random plans by jer
 *  
 *  ok based on gprof, we really need some innovation here... my thoughs are this:
 *  
 *  most things are strings, so have a string-based true-blue garbage collector
 *  one big global hash containing all the strings created by any pstrdup, returning const char *
 *  a refcount on each string block
 *  when a pool is freed, it moves down the refcount
 *  garbage collector collects pools on the free stack, and runs through the hash for unused strings
 *  j_strcmp can check for == (if they are both from a pstrdup)
 *  
 *  let's see... this would change:
 *  pstrdup: do a hash lookup, success=return, fail=pmalloc & hash put
 *  pool_free: 
 *  
 *  
 *  
 *  
 *  
 */

#include "libxode.h"
#include "config.h"

/********************************************/
/* DEBUG functions first, normal ones later */
/********************************************/
#ifdef POOLDEBUG 

pool _pools = NULL;
int _poolblocks = 0;
int _poolstat = 0;

pool pool_new_d(char *zone)
{
    pool p;
    p = malloc(sizeof(_pool));
    p->size = p->debug = 0;
    p->stat = p->cstat = _poolstat;
    p->label = strdup(zone);
    p->next = NULL;
    p->block = (void *)_pools;
    _pools = p;
    /*    fprintf(stderr,"_ _ _ _ / %X %s:%d \\ _ _ _ _\n",p,p->file,p->line);*/
    return p;
}

void *pmalloc_d(char *zone, pool p, int size)
{
    void *block = NULL;
    pool new;

    block = malloc(size);

    if(p == NULL) /* this is most likely an error that we got a NULL here! */
    {
        fprintf(stderr,"Memory Leak! [pmalloc received NULL pool, unable to track allocation]");
    }else{
        new = malloc(sizeof(_pool));
        new->size = size;
        new->block = block;
        new->next = p->next;
        new->label = strdup(zone);
        new->stat = p->cstat = _poolstat;
        p->size += size;
        _poolblocks += size;
        p->next = new;
        if(p->debug)
            fprintf(stderr,"_ _ _ _ / %X +%d [%s/%s] \\ _ _ _ _\n",p,size,p->label,zone);
    }

    return block;
}

char *pstrdup_d(char *zone, pool p, const char *src)
{
    char *ret;

    if(src == NULL)
        return NULL;

    ret = pmalloc_d(zone, p,strlen(src) + 1);
    strcpy(ret,src);

    return ret;
}

int pool_size(pool p)
{
    if(p == NULL) return 0;

    return p->size;
}

void pool_free(pool p)
{
    pool next, current;

    if(p == NULL)
        return;

    /* unlink p from the master list */
    if(p == _pools)
        _pools = (pool)(_pools->block);
    else
        for(current = _pools; current != NULL; current = (pool)(current->block))
            if(p == (pool)(current->block))
                current->block = (pool)(p->block);

    if(p->debug)
        fprintf(stderr,"_ _ _ _ \\ %X -%d [%s] / _ _ _ _\n",p,p->size,p->label);
    p->size = 0;
    p->block = NULL;
    current = p;
    while(current != NULL)
    {
        next = current->next;
        _poolblocks -= current->size;
        free(current->block);
        free(current->label);
        free(current);
        current = next;
    }
}

void pool_stat(int full)
{
    pool cur, cur2;

    fprintf(stderr,"_ _ _ _ > POOL %d < _ _ _ _\n",_poolblocks);
    for(cur = _pools; cur != NULL; cur = (pool)(cur->block))
    {
        if(cur->stat == _poolstat)
            fprintf(stderr,"%X %d (new)%s\n",cur,cur->size,cur->label);
        else if(full || cur->cstat == _poolstat)
            fprintf(stderr,"%X %d %s\n",cur,cur->size,cur->label);

        continue;
        if(cur->stat != _poolstat && cur->cstat != _poolstat) continue;
        if(cur->cstat)
            for(cur2 = cur->next; cur2 != NULL; cur2 = cur2->next)
                if(cur2->stat == _poolstat)
                    fprintf(stderr,"        +%d %s\n",cur2->size,cur2->label);
    }
    fprintf(stderr,"_ _ _ _ _ _ _ _ _ _ _ _ _ _ _\n");
    ++_poolstat;
}

void pool_label(pool p, char *label, int debug)
{
    if(p == NULL || label == NULL) return;

    free(p->label);
    p->label = strdup(label);
    p->debug = debug;
}

/********************************************/
#else
/********************************************/
int total = 0;

/*
void *_pool__malloc(size_t size)
{
    total++;
    return malloc(size);
}

void _pool__free(void *block)
{
    total--;
    free(block);
}
*/

#define _pool__malloc malloc
#define _pool__free free

/* make an empty pool */
pool pool_new(void)
{
    pool p;
    p = _pool__malloc(sizeof(_pool));
    p->cleanup = NULL;
    p->heap = NULL;
    p->size = 0;
    return p;
}

/* free a heap */
void _pool_heap_free(void *arg)
{
    struct pheap *h = (struct pheap *)arg;

    /* total -= h->size;
    printf("freeing heap %X size %d total %d\n",h->block,h->size,total);*/
    _pool__free(h->block);
    _pool__free(h);
}

/* mem should always be freed last */
void _pool_cleanup_append(pool p, struct pfree *pf)
{
    struct pfree *cur;

    if(p->cleanup == NULL)
    {
        p->cleanup = pf;
        return;
    }

    /* fast forward to end of list */
    for(cur = p->cleanup; cur->next != NULL; cur = cur->next);

    cur->next = pf;
}

/* create a cleanup tracker */
struct pfree *_pool_free(pool p, pool_cleaner f, void *arg)
{
    struct pfree *ret;

    /* make the storage for the tracker */
    ret = _pool__malloc(sizeof(struct pfree));
    ret->f = f;
    ret->arg = arg;
    ret->next = NULL;

    return ret;
}

/* create a heap and make sure it get's cleaned up */
struct pheap *_pool_heap(pool p, int size)
{
    struct pheap *ret;
    struct pfree *clean;

    /* make the return heap */
    ret = _pool__malloc(sizeof(struct pheap));
    ret->block = _pool__malloc(size);
    ret->size = size;
    p->size += size;
    ret->used = 0;

    /* total += size;
    printf("creating heap %X size %d total %d\n",ret->block,size,total);*/

    /* append to the cleanup list */
    clean = _pool_free(p, _pool_heap_free, (void *)ret);
    clean->heap = ret; /* for future use in finding used mem for pstrdup */
    _pool_cleanup_append(p, clean);

    return ret;
}

pool pool_heap(int size)
{
    pool p;
    p = pool_new();
    p->heap = _pool_heap(p,size);
    return p;
}

void *pmalloc(pool p, int size)
{
    void *block;

    if(p == NULL)
    {
        fprintf(stderr,"Memory Leak! [pmalloc received NULL pool, unable to track allocation, exiting]\n");
        abort();
    }

    /* if there is no heap for this pool or it's a big request, just raw, I like how we clean this :) */
    if(p->heap == NULL || size > (p->heap->size / 2))
    {
        block = _pool__malloc(size);
        p->size += size;
        _pool_cleanup_append(p, _pool_free(p, _pool__free, block));
        return block;
    }

    /* we have to preserve boundaries, long story :) */
    if(size >= 4)
        while(p->heap->used&7) p->heap->used++;

    /* if we don't fit in the old heap, replace it */
    if(size > (p->heap->size - p->heap->used))
        p->heap = _pool_heap(p, p->heap->size);

    /* the current heap has room */
    block = (char *)p->heap->block + p->heap->used;
    p->heap->used += size;
    return block;
}

void *pmalloc_x(pool p, int size, char c)
{
   void* result = pmalloc(p, size);
   if (result != NULL)
           memset(result, c, size);
   return result;
}  

/* easy safety utility (for creating blank mem for structs, etc) */
void *pmalloco(pool p, int size)
{
    void *block = pmalloc(p, size);
    memset(block, 0, size);
    return block;
}  

/* XXX efficient: move this to const char * and then loop throug the existing heaps to see if src is within a block in this pool */
char *pstrdup(pool p, const char *src)
{
    char *ret;

    if(src == NULL)
        return NULL;

    ret = pmalloc(p,strlen(src) + 1);
    strcpy(ret,src);

    return ret;
}

/* when move above, this one would actually return a new block */
char *pstrdupx(pool p, const char *src)
{
    return pstrdup(p, src);
}

int pool_size(pool p)
{
    if(p == NULL) return 0;

    return p->size;
}

void pool_free(pool p)
{
    struct pfree *cur, *stub;

    cur = p->cleanup;
    while(cur != NULL)
    {
        (*cur->f)(cur->arg);
        stub = cur->next;
        _pool__free(cur);
        cur = stub;
    }

    _pool__free(p);

/*printf("\n\ntotal missed mallocs %d\n\n",total);*/
}

/* public cleanup utils, insert in a way that they are run FIFO, before mem frees */
void pool_cleanup(pool p, pool_cleaner f, void *arg)
{
    struct pfree *clean;

    clean = _pool_free(p, f, arg);
    clean->next = p->cleanup;
    p->cleanup = clean;
}

void pool_label(pool p, char *label, int debug)
{
    return;
}

void pool_stat(int full)
{
    return;
}

/********************************************/
#endif /* POOLDEBUG */
/********************************************/
